/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%											
%   "Copyright (c) 2004, 2005 The University of Southern California"				
%   All rights reserved.								
%											
%   Permission to use, copy, modify, and distribute this script and its		
%   documentation for any purpose, without fee, and without written agreement is	
%   hereby granted, provided that the above copyright notice, the following		
%   two paragraphs and the names in the credits appear in all copies of this software.		
%											
%   NO REPRESENTATIONS ARE MADE ABOUT THE SUITABILITY OF THE SCRIPT FOR ANY		
%   PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY. NO 
%   LIABILITY IS ASSUMED BY THE DEVELOPERS.
%											
%											
%   Authors:  	     Kiran Yedavalli
%   Director:        Prof. Bhaskar Krishnamachari
%   Autonomous Networks Research Group, University of Southern California
%   http://ceng.usc.edu/~anrg
%
%   Contact: kyedavalli@gmail.com
%   Current Version: 1.0, 2005/11/01
%   Date last modified: 2005/11/01							
%											
%   Anything following a // is treated as a comment.					
%											
%											
%   Description:									
%	This program evaluates the location of unknown nodes using
%  	the Sequence-Based Localization technique. It also impliments
%  	Proximity, Maximum Likelihood Estimation (Least Squares Estimation),
%  	and Centroid techniques for comparison.
%	
%	This technique is described in detail in:
%  	Kiran Yedavalli, Bhaskar Krishnamachari, "Sequence-Based Localization in Wireless Sensor Networks",
%  	To Appear in IEEE Transaction on Mobile Computing, 2007.
%
%  	We used the g++ compiler to compile the program.
%											
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/



#include "sbl.h"

int main(int argc, char **argv){
  
  if(argc == 1){
    printIt = 0;
  }
  else if(argc == 2){
    if(!strcmp(*(argv+1),"-d")){
      printIt = 1;
    }
    else if(!strcmp(*(argv+1),"-l")){
      // print anchors and face list
      printIt = 2;
    }
    else if(!strcmp(*(argv+1),"-sq")){
      // print the sequence list
      printIt = 3;
    }
    else if(!strcmp(*(argv+1),"-s")){
      // print the spearman's correlations
      printIt = 4;
    }
  }
  
  // Variable declerations
  // Lists
  FaceList *FL;                // List to store faces in the arrangement of lines.
  EdgeList *EL;                // List to store edges in the arrangement of lines.
  VertexList *VL;              // List to store vertices in the arrangement of lines.  
  
  // Bounding box variables
  HalfEdge *be[4];
  Vertex *bv[4];

  // Anchors (Reference nodes) and lines
  Line *line[LINES];
  double m,c;
  int numLines,count,check;
  
  // Other Variables
  HalfEdge *e;
  Vertex *v;
  Face *f;
  HalfEdge *temp,*presentEdge,*tempEdge;
  int leaving;   // Is the new line leaving or entering the edge 
  int DONE;
  int rssRankSequence[ANCHORS],rank[ANCHORS];
  float maxRss,T,d[ANCHORS],minD;

  // Kendall Variables
  float kx,ky,kendall[MAXNUM],maxKendall,kError;
  int numMaxKendall,nc,nd,ntx,nty;

  // mle variables
  float mLV[125][125],minLV,mlex,mley,mleError,gx,gy,md[ANCHORS],dist;
  int numMin;
  
  // Proximity variables
  float prox,proy,proError;
  int numPro;
  
  // Centroid variables
  float cenx,ceny,cenError;
  int numCen;

  /*

  This program uses values from an experiment conducted using MICA 2 motes
  in a parking lot at USC. In this experiment the unknown node locations have been 
  chosen to be the locations of the anchors and compared with the true locations
  to evaluate the localization technique.

  */

  // Variable to point to the current unknown node location that is being estimated.
  // Its value ranges from 0 to ANCHORS
  int node = 1;  

  // Unknown node coordinates in feet (true locations) 
  float tx[ANCHORS+1] = {1.0,14.0,3.0,21.0,9.0,22.0,10.0,3.0,27.0,20.0,18.0};
  float ty[ANCHORS+1] = {7.0,3.0,19.0,18.0,33.0,3.0,11.0,32.0,27.0,12.0,34.0};
  
  // Unknown node coordinates in meters
  for(int i=0;i<ANCHORS+1;i++){
    tx[i] = tx[i]/(3.0);
    ty[i] = ty[i]/(3.0);
  }

  // Anchor coordinates in feet
  float allx[ANCHORS+1][ANCHORS] = {{14.0,3.0,21.0,9.0,22.0,10.0,3.0,27.0,20.0,18.0},
				 {1.0,3.0,21.0,9.0,22.0,10.0,3.0,27.0,20.0,18.0},
				 {1.0,14.0,21.0,9.0,22.0,10.0,3.0,27.0,20.0,18.0},
				 {1.0,14.0,3.0,9.0,22.0,10.0,3.0,27.0,20.0,18.0},
				 {1.0,14.0,3.0,21.0,22.0,10.0,3.0,27.0,20.0,18.0},
				 {1.0,14.0,3.0,21.0,9.0,10.0,3.0,27.0,20.0,18.0},
				 {1.0,14.0,3.0,21.0,9.0,22.0,3.0,27.0,20.0,18.0},
				 {1.0,14.0,3.0,21.0,9.0,22.0,10.0,27.0,20.0,18.0},
				 {1.0,14.0,3.0,21.0,9.0,22.0,10.0,3.0,20.0,18.0},
				 {1.0,14.0,3.0,21.0,9.0,22.0,10.0,3.0,27.0,18.0},
				 {1.0,14.0,3.0,21.0,9.0,22.0,10.0,3.0,27.0,20.0}};
  
  
  float ally[ANCHORS+1][ANCHORS] = {{3.0,19.0,18.0,33.0,3.0,11.0,32.0,27.0,12.0,34.0},
				 {7.0,19.0,18.0,33.0,3.0,11.0,32.0,27.0,12.0,34.0},
				 {7.0,3.0,18.0,33.0,3.0,11.0,32.0,27.0,12.0,34.0},
				 {7.0,3.0,19.0,33.0,3.0,11.0,32.0,27.0,12.0,34.0},
				 {7.0,3.0,19.0,18.0,3.0,11.0,32.0,27.0,12.0,34.0},
				 {7.0,3.0,19.0,18.0,33.0,11.0,32.0,27.0,12.0,34.0},
				 {7.0,3.0,19.0,18.0,33.0,3.0,32.0,27.0,12.0,34.0},
				 {7.0,3.0,19.0,18.0,33.0,3.0,11.0,27.0,12.0,34.0},
				 {7.0,3.0,19.0,18.0,33.0,3.0,11.0,32.0,12.0,34.0},
				 {7.0,3.0,19.0,18.0,33.0,3.0,11.0,32.0,27.0,34.0},
				 {7.0,3.0,19.0,18.0,33.0,3.0,11.0,32.0,27.0,12.0}};
  
  // Anchor coordinates in meters
  for(int i=0;i<ANCHORS;i++){
    ax[i] = allx[node][i]/3.0;
    ay[i] = ally[node][i]/3.0;
  }

  // Recieved signal strength readings from the experiment in dBm.
  float rss[ANCHORS+1][ANCHORS] = {{-68.308,-62.299,-67.051,-68.141,-66.400,-60.864,-77.891,-68.217,-69.553,-68.678},
				   {-67.414,-68.205,-65.299,-69.623,-55.270,-59.656,-71.891,-69.178,-65.439,-74.507},
				   {-60.025,-67.713,-65.146,-64.812,-67.982,-58.420,-62.949,-70.291,-68.334,-69.354},
				   {-68.820,-67.432,-67.988,-68.111,-64.830,-65.680,-70.307,-56.207,-53.938,-67.278},
				   {-66.090,-68.082,-64.801,-64.279,-73.314,-63.975,-49.990,-62.357,-71.328,-54.147},
				   {-67.572,-55.854,-70.355,-64.174,-77.715,-77.591,-88.658,-68.432,-57.822,-73.303},
				   {-62.527,-62.622,-62.680,-65.855,-68.123,-70.461,-70.402,-73.127,-65.211,-79.361},
				   {-77.521,-74.398,-65.662,-70.068,-50.838,-86.564,-70.062,-70.560,-81.057,-67.722},
				   {-68.896,-70.385,-72.194,-54.641,-63.342,-67.637,-70.941,-69.543,-65.686,-63.720},
				   {-66.488,-63.775,-66.978,-51.330,-69.295,-56.240,-63.301,-78.221,-65.041,-70.042},
				   {-72.078,-80.838,-73.341,-68.369,-57.963,-76.385,-79.549,-68.691,-66.980,-73.718}};
  
  // Measured distance between the anchors and the unknown node,
  // used by the MLE (LSE) technique.
  for(int i=0;i<ANCHORS;i++)
    md[i] = pow(10.0,((TXP - PL - rss[node][i])/(ETA*10.0)));
  
  // Unknown node location sequence.
  // Rank the anchors based on RSS values
  count = 0;
  while(count < ANCHORS){
    maxRss = -pow(10.0,20.0);
    for(int p=0;p<ANCHORS;p++){
      if(rss[node][p] > maxRss)
	maxRss = rss[node][p];
    }
    check = 0;
    for(int p=0;p<ANCHORS;p++){
      if(rss[node][p] == maxRss){
	rss[node][p] = -pow(10.0,20.0);
	rssRankSequence[p] = count;
	check++;
      }
    }
    count = count + check;
  }

  // Get the T for the true location
  for(int i=0;i<ANCHORS;i++)
    d[i] = sqrt(pow((tx[node] - ax[i]),2)+pow((ty[node] - ay[i]),2));
  
  count = 0;
  while(count < ANCHORS){
    minD = pow(10.0,20.0);
    for(int p=0;p<ANCHORS;p++){
      if(d[p] < minD)
	minD = d[p];
    }
    check = 0;
    for(int p=0;p<ANCHORS;p++){
      if(d[p] == minD){
	d[p] = pow(10.0,20.0);
	rank[p] = count;
	check++;
      }
    }
    count = count + check;
  }
  
  nc = nd = ntx = nty = 0;
  for(int i=0;i<ANCHORS;i++){
    for(int j=i+1;j<ANCHORS;j++){
      if(((rank[i] > rank[j]) && (rssRankSequence[i] > rssRankSequence[j])) || 
	 ((rank[i] < rank[j]) && (rssRankSequence[i] < rssRankSequence[j]))) 
	nc++;
      else if(((rank[i] > rank[j]) && (rssRankSequence[i] < rssRankSequence[j])) || 
	      ((rank[i] < rank[j]) && (rssRankSequence[i] > rssRankSequence[j]))) 
	nd++;
      else if(rank[i] == rank[j])
	ntx++;
      else if(rssRankSequence[i] == rssRankSequence[j])
	nty++;
    }
  }
  T = (nc-nd)/(sqrt(nc+nd+ntx)*sqrt(nc+nd+nty));
  
  EL = new EdgeList;
  FL = new FaceList;
  VL = new VertexList;
  
  // The Bounding Box 
  if(printIt == 1){
    printf("\n*** THE BOUNDING BOX ***\n");
  }
  
  // Vertices
  bv[0] = new Vertex(RANGE*1.0,RANGE*1.0);
  bv[1] = new Vertex(0.0,RANGE*1.0);
  bv[2] = new Vertex(0.0,0.0);
  bv[3] = new Vertex(RANGE*1.0,0.0);
  
  // HalfEdges
  for(int i=0;i<4;i++){
    be[i] = new HalfEdge(bv[i],bv[((i+1)%4)]);
  }
  
  // Next and Prev HalfEdges
  for(int i=0;i<4;i++){
    be[i]->next = be[((i+1)%4)];
    be[i]->prev = be[((i+3)%4)];
    be[i]->twin = NULL;
  }
  
  // Create new face and add it to the facelist
  f = new Face;
  f->incEdge = be[0];
  FL->addFace(f);
  
  // Update edges and add edges to EdgeList
  for(int i=0;i<4;i++){
    be[i]->incFace = f;
    EL->addEdge(be[i]);
  }
  
  // Update vertices and add them to VertexList
  for(int i=0;i<4;i++){
    bv[i]->incEdge = be[i];
    VL->addVertex(bv[i]);
  }
  
  // ANCHORS
  if(printIt == 1){
    printf("\n*** ANCHORS ***\n");
  }
  
  // Lines
  numLines = 0;
  for(int i=0;i<ANCHORS;i++){
    for(int j=i+1;j<ANCHORS;j++){

      line[numLines] = new Line;
      
      if(fabs(ay[i] - ay[j]) == 0){
	// The line is verticle
	line[numLines]->lx = (ax[i] + ax[j])*0.5;
	line[numLines]->ly = RANGE*1.0;
	
	line[numLines]->rx = (ax[i] + ax[j])*0.5;
	line[numLines]->ry = 0.0;
      }
      else{
	m = -(ax[i]-ax[j])/(ay[i]-ay[j]);
	c = (ay[i] + ay[j])*0.5 - m*(ax[i] + ax[j])*0.5;  
	
	if(c < 0){
	  line[numLines]->lx = -(c*1.0)/(m*1.0);
	  line[numLines]->ly = 0.0;
	  if((m*RANGE*1.0+c) > RANGE*1.0){
	    line[numLines]->rx = (RANGE*1.0 - c)/(m*1.0);
	    line[numLines]->ry = RANGE*1.0;
	  }
	  else{
	    line[numLines]->rx = RANGE*1.0;
	    line[numLines]->ry = m*RANGE*1.0+c;
	  }
	}
	else if((c >= 0.0) && (c <= RANGE*1.0)){
	  line[numLines]->lx = 0.0; 
	  line[numLines]->ly = c;
	  if(m*RANGE*1.0+c < 0.0){
	    line[numLines]->rx = -(c*1.0)/(m*1.0);
	    line[numLines]->ry = 0.0;
	  }
	  else if((m*RANGE*1.0+c >= 0.0) && (m*RANGE*1.0+c <= RANGE*1.0)){
	    line[numLines]->rx = RANGE*1.0;
	    line[numLines]->ry = m*RANGE*1.0+c;
	  }
	  else{
	    line[numLines]->rx = (RANGE*1.0-c)/(m*1.0);
	    line[numLines]->ry = RANGE*1.0;
	  }
	}
	else{
	  line[numLines]->lx = (RANGE*1.0-c*1.0)/(m*1.0);
	  line[numLines]->ly = RANGE*1.0;
	  if(m*RANGE*1.0+c < 0.0){
	    line[numLines]->rx = -(c*1.0)/(m*1.0);
	    line[numLines]->ry = 0.0;
	  }
	  else{
	    line[numLines]->rx = RANGE*1.0;
	    line[numLines]->ry = m*RANGE*1.0+c;
	  }
	}	
      }
      check = 0;
      for(int k=0;k<numLines;k++){
	if((line[k]->lx == line[numLines]->lx) && (line[k]->ly == line[numLines]->ly)
	   && (line[k]->rx == line[numLines]->rx) && (line[k]->ry == line[numLines]->ry))
	  check++;
      }
      if(check == 0)
	numLines++;
    }
  }
  
  if(printIt == 1){
    printf("\n*** LINES ***\n");
    for(int i=0;i<numLines;i++){
      printf("%d\t[%f, %f]->[%f, %f]\n",i,line[i]->lx,line[i]->ly,line[i]->rx,line[i]->ry);
    }
  }
  
  // Adding lines to the bounding box 
  for(int i=0;i<numLines;i++){
    
    // Adding a new line
    if(printIt == 1){
      printf("\n*** ADDING LINE %d: ([%f, %f]->[%f, %f])***\n",i,line[i]->lx,line[i]->ly,
	     line[i]->rx,line[i]->ry);
    }
    
    // The new line always leaves the edge on the bounding box
    leaving = 1;
    
    // Get the edge on the boundary that intersects the line 
    if(VL->isPointNew(line[i]->lx,line[i]->ly)){
      presentEdge = EL->getEdgeWithPoint(line[i]->lx,line[i]->ly);
    }
    else{
      presentEdge = EL->getBoundaryEdgeWithOrig(line[i]->lx,line[i]->ly);
    }
    
    DONE = 0;
    while(!DONE){
      
      if(printIt == 1){
	printf("\nLINE: %d\tPRESENT-EDGE: ([%f, %f]->[%f, %f])\n",i,
	       presentEdge->orig->x,presentEdge->orig->y,presentEdge->dest->x,presentEdge->dest->y);
      }
      
      // Do line and presentEdge intersect
      if(isThereIntersection(line[i],presentEdge)){
	
	if(leaving){
	  
	  // The new line is leaving the edge
	  if(printIt == 1){
	    printf("The New Line is Leaving The Edge\n");
	  }
	  // Get the intersection vertex
	  v = getVertex(line[i],presentEdge);
	  
	  if(printIt == 1){
	    printf("Intersection: (%f, %f)\n",v->x,v->y);
	  }
	  if(isVertexNewToEdge(presentEdge,v) == 2){
	    
	    // Its a new vertex
	    // Split the intersecting edge
	    // Create two new edges in place of old edge
	    e = new HalfEdge(presentEdge->orig,v);
	    e->prev = presentEdge->prev;
	    presentEdge->prev->next = e;
	    if(EL->isThereEdgeWithOrigDest(v,presentEdge->orig)){
	      e->twin = EL->getEdgeWithOrigDest(v,presentEdge->orig);
	      EL->getEdgeWithOrigDest(v,presentEdge->orig)->twin = e;
	    }
	    else{
	      e->twin = NULL;
	    }
	    EL->addEdge(e);
	    
	    e = new HalfEdge(v,presentEdge->dest);
	    e->next = presentEdge->next;
	    presentEdge->next->prev = e;
	    if(EL->isThereEdgeWithOrigDest(presentEdge->dest,v)){
	      e->twin = EL->getEdgeWithOrigDest(presentEdge->dest,v);
	      EL->getEdgeWithOrigDest(presentEdge->dest,v)->twin = e;
	    }
	    else{
	      e->twin = NULL;
	    }
	    v->incEdge = e;
	    EL->addEdge(e);
	    
	    // Store the intersected edge, update the number 
	    // and remove it from the edgelist
	    line[i]->le[line[i]->ne] = EL->getEdgeWithOrigDest(presentEdge->orig,v);
	    line[i]->ne++;
	    line[i]->le[line[i]->ne] = EL->getEdgeWithOrigDest(v,presentEdge->dest);
	    line[i]->ne++;
	    EL->deleteEdge(presentEdge);
	    
	    // Store the vertex, update its number and add
	    // it to the vertex list
	    line[i]->lv[line[i]->nv] = v;
	    line[i]->nv++;
	    if(VL->isPointNew(v->x,v->y))
	      VL->addVertex(v);
	  }
	  else if(isVertexNewToEdge(presentEdge,v) == 1){
	    // vertex is the destination of the edge

	    // Store the present edge and update its number
	    line[i]->le[line[i]->ne] = EL->getEdgeWithOrigDest(presentEdge->orig,v);
	    line[i]->ne++;
	    line[i]->le[line[i]->ne] = EL->getEdgeWithOrigDest(v,presentEdge->next->dest);
	    line[i]->ne++;
	    
	    //Store the vertex, update its number
	    line[i]->lv[line[i]->nv] = v;
	    line[i]->nv++;
	  }
	  else{
	    // vertex is the origin of the edge
	    
	    // Store the prev edge of the present edge and update its number
	    line[i]->le[line[i]->ne] = EL->getEdgeWithOrigDest(presentEdge->prev->orig,v);
	    line[i]->ne++;
	    line[i]->le[line[i]->ne] = EL->getEdgeWithOrigDest(v,presentEdge->dest);
	    line[i]->ne++;
	    
	    //Store the vertex, update its number
	    line[i]->lv[line[i]->nv] = v;
	    line[i]->nv++;
	    
	  }
	  // What ever the case proceed to the next halfedge 
	  leaving = 0;
	  if(presentEdge->next != NULL)
	    presentEdge = presentEdge->next;
	  else
	    DONE = 1;
	}
	else{
	  // The new line is entering the edge
	  if(printIt == 1){
	    printf("The New Line is Entering The Edge\n");
	  }
	  // Get the vertex of intersection
	  v = getVertex(line[i],presentEdge);
	  if(printIt == 1){
	    printf("Intersection: (%f, %f)\n",v->x,v->y);
	  }
	  if(isVertexNewToEdge(presentEdge,v) == 2){
	    
	    // Remove the old face from the face list
	    FL->deleteFace(presentEdge->incFace);
	    
	    // Create new edges for the new line segment
	    // and new faces in place of old face
	    e = new HalfEdge(v,line[i]->lv[line[i]->nv-1]);
	    e->next = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->le[line[i]->ne-1]->dest);
	    EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->le[line[i]->ne-1]->dest)->prev = e;
	    f = new Face;
	    f->incEdge = e;
	    e->incFace = f;
	    FL->addFace(f);
	    EL->addEdge(e);
	    
	    e = new HalfEdge(line[i]->lv[line[i]->nv-1],v);
	    e->prev = EL->getEdgeWithOrigDest(line[i]->le[line[i]->ne-2]->orig,line[i]->lv[line[i]->nv-1]);
	    EL->getEdgeWithOrigDest(line[i]->le[line[i]->ne-2]->orig,line[i]->lv[line[i]->nv-1])->next = e;
	    e->twin = EL->getEdgeWithOrigDest(v,line[i]->lv[line[i]->nv-1]);
	    EL->getEdgeWithOrigDest(v,line[i]->lv[line[i]->nv-1])->twin = e;
	    f = new Face;
	    f->incEdge = e;
	    e->incFace = f;
	    FL->addFace(f);
	    EL->addEdge(e);
	    
	    // Split the intersected edge
	    // Create new edges for the intersected edge
	    e = new HalfEdge(presentEdge->orig,v);
	    e->prev = presentEdge->prev;
	    presentEdge->prev->next = e;
	    e->next = EL->getEdgeWithOrigDest(v,line[i]->lv[line[i]->nv-1]);
	    EL->getEdgeWithOrigDest(v,line[i]->lv[line[i]->nv-1])->prev = e;
	    EL->addEdge(e);
	    
	    e = new HalfEdge(v,presentEdge->dest);
	    e->prev = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],v);
	    EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],v)->next = e;
	    e->next = presentEdge->next;
	    presentEdge->next->prev = e;
	    v->incEdge = e;
	    EL->addEdge(e);
	    
	    // Remove presentEdge from the EdgeList
	    EL->deleteEdge(presentEdge);
	    
	    // Store the new vertex, update its number and
	    // add it to the vertex list
	    line[i]->lv[line[i]->nv] = v;
	    line[i]->nv++;
	    if(VL->isPointNew(v->x,v->y))
	      VL->addVertex(v);
	    
	    // Associate faces to all edges as new faces have been created
	    if(printIt == 1){
	      printf("Starting To Associate Faces To Edges\n");
	    }
	    
	    temp = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2]);
	    while(temp->next != EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2])){
	      temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2])->incFace;
	      temp = temp->next;
	    }
	    temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2])->incFace;
	    
	    temp = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1]);
	    while(temp->next != EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1])){
	      temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1])->incFace;
	      temp = temp->next;
	    }
	    temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1])->incFace;
	    if(printIt == 1){
	      printf("Done Associating Faces To Edges\n");
	    }
	  }
	  else{
	    // Its an old Vertex
	    
	    if(isVertexNewToLine(line[i],v)){
	      
	      // Remove the old face from the face list
	      FL->deleteFace(presentEdge->incFace);
	      
	      // Temporarily store the next of presentEdge
	      tempEdge = presentEdge->next;
	      
	      // Create new edges for the new line segment
	      // and new faces in place of old face
	      e = new HalfEdge(v,line[i]->lv[line[i]->nv-1]);
	      e->next = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->le[line[i]->ne-1]->dest);
	      EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->le[line[i]->ne-1]->dest)->prev = e;
	      e->prev = presentEdge;
	      presentEdge->next = e;
	      f = new Face;
	      f->incEdge = e;
	      e->incFace = f;
	      FL->addFace(f);
	      EL->addEdge(e);
	      
	      e = new HalfEdge(line[i]->lv[line[i]->nv-1],v);
	      e->prev = EL->getEdgeWithOrigDest(line[i]->le[line[i]->ne-2]->orig,line[i]->lv[line[i]->nv-1]);
	      EL->getEdgeWithOrigDest(line[i]->le[line[i]->ne-2]->orig,line[i]->lv[line[i]->nv-1])->next = e;
	      e->next = tempEdge;
	      tempEdge->prev = e;
	      e->twin = EL->getEdgeWithOrigDest(v,line[i]->lv[line[i]->nv-1]);
	      EL->getEdgeWithOrigDest(v,line[i]->lv[line[i]->nv-1])->twin = e;
	      f = new Face;
	      f->incEdge = e;
	      e->incFace = f;
	      FL->addFace(f);
	      EL->addEdge(e);
	      
	      // Store the new vertex 
	      line[i]->lv[line[i]->nv] = v;
	      line[i]->nv++;
	      
	      // Associate faces to all edges as new faces have been created
	      if(printIt == 1){
		printf("Starting to Associate Faces To Edges\n");
	      }
	      temp = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2]);
	      while(temp->next != EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2])){
		temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2])->incFace;
		temp = temp->next;
	      }
	      temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-1],line[i]->lv[line[i]->nv-2])->incFace;
	      
	      temp = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1]);
	      while(temp->next != EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1])){
		temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1])->incFace;
		temp = temp->next;
	      }
	      temp->incFace = EL->getEdgeWithOrigDest(line[i]->lv[line[i]->nv-2],line[i]->lv[line[i]->nv-1])->incFace;
	      if(printIt == 1){
		printf("Done Associating Faces To Edges\n");
	      }
	    }
	    else{
	      // Vertex is not new to the line
	      // DO Nothing
	    }
	  }
	  
	  // Proceed to the next face whatever the case
	  leaving = 1;
	  if(EL->isThereEdgeWithOrigDest(presentEdge->dest,presentEdge->orig)){
	    presentEdge = EL->getEdgeWithOrigDest(presentEdge->dest,presentEdge->orig);
	  }else{
	    DONE = 1;
	  }
	}     
      }
      else{ 
	// If they dont intersect 
	if(presentEdge->next != NULL){
	  presentEdge = presentEdge->next;
	}else{
	  DONE = 1;
	}
      }
      
      if(printIt == 1){
	printf("Stored Edges In Line %d.\n",i);
	for(int j=0;j<line[i]->ne;j++){
	  printf("%d\t[%f, %f]->[%f, %f]\n",j,line[i]->le[j]->orig->x,line[i]->le[j]->orig->y,
		 line[i]->le[j]->dest->x,line[i]->le[j]->dest->y);
	}
	printf("Stored Vertices In Line %d.\n",i);
	for(int j=0;j<line[i]->nv;j++){
	  printf("%d\t[%f, %f]\n",j,line[i]->lv[j]->x,line[i]->lv[j]->y);
	}
      }
      
      if(!DONE){
	if(printIt == 1){
	  printf("Going To The Next Half-Edge: ([%f, %f]->[%f, %f])\n",
		 presentEdge->orig->x,presentEdge->orig->y,presentEdge->dest->x,presentEdge->dest->y);
	}
      }
      else{
	if(printIt == 1){
	  printf("*** THE END ***\n");
	}
      }
    }
    if(printIt == 1){
      VL->printVertexList();
      EL->printEdgeList();
      FL->printFaceList();
    }   
    
    if(printIt == 1){
      printf("Done With Line %d.\n",i);
    }
  }
  
  FL->getRankSequences();
  EL->getRankSequences();
  VL->getRankSequences();
  
  if(printIt == 3){
    for(int i=0;i<numRankSequences;i++){
      printf("%d\t",i);
      for(int j=0;j<ANCHORS;j++){
	printf("%d",rankSequence[i][j]);
      }
      printf("\t[%f, %f]\n",centroidList[i].cx,centroidList[i].cy);
    }
  }

  // Search the sequence list
  maxKendall = -pow(10.0,20.0);
  for(int i=0;i<numRankSequences;i++){
    nc = nd = ntx = nty = 0;
    for(int j=0;j<ANCHORS;j++){
      for(int k=j+1;k<ANCHORS;k++){
	if(((rankSequence[i][j] > rankSequence[i][k]) && (rssRankSequence[j] > rssRankSequence[k])) ||
	   ((rankSequence[i][j] < rankSequence[i][k]) && (rssRankSequence[j] < rssRankSequence[k])))
	  nc++;
	else if(((rankSequence[i][j] > rankSequence[i][k]) && (rssRankSequence[j] < rssRankSequence[k])) ||
		((rankSequence[i][j] < rankSequence[i][k]) && (rssRankSequence[j] > rssRankSequence[k])))
	  nd++;
	else if(rankSequence[i][j] == rankSequence[i][k])
	  ntx++;
	else if(rssRankSequence[j] == rssRankSequence[k])
	  nty++;
      }
    }
    kendall[i] = ((nc-nd)*1.0)/(sqrt(nc+nd+ntx)*sqrt(nc+nd+nty));
    if(kendall[i] > maxKendall)
      maxKendall = kendall[i];
  }
  
  kx = ky = 0.0;
  numMaxKendall = 0;
  for(int i=0;i<numRankSequences;i++){
    if(kendall[i] == maxKendall){
      kx += centroidList[i].cx;
      ky += centroidList[i].cy;
      numMaxKendall++;
    }
  }
  kx = kx/(numMaxKendall*1.0);
  ky = ky/(numMaxKendall*1.0);
  kError = sqrt(pow((kx - tx[node]),2)+pow((ky - ty[node]),2));

  // mle
  minLV = pow(10.0,20.0);
  for(int i=0;i<(RANGE*RES)+1;i++){
    for(int j=0;j<(RANGE*RES)+1;j++){
      
      gx = (i*1.0)/(RES*1.0);
      gy = (j*1.0)/(RES*1.0);

      mLV[i][j] = 0.0;
      for(int k=0;k<ANCHORS;k++){
	dist = sqrt(pow((gx - ax[k]),2)+pow((gy - ay[k]),2));
	mLV[i][j] += pow((dist-md[k]),2);
      }

      if(mLV[i][j] < minLV)
	minLV = mLV[i][j];
    }
  }
  mlex = mley = 0.0;
  numMin = 0;
  for(int i=0;i<(RANGE*RES)+1;i++){
    for(int j=0;j<(RANGE*RES)+1;j++){
      gx = (i*1.0)/(RES*1.0);
      gy = (j*1.0)/(RES*1.0);
      if(mLV[i][j] == minLV){
	mlex += gx;
	mley += gy;
	numMin++;
      }
    }
  }
  mlex = mlex/(numMin*1.0);
  mley = mley/(numMin*1.0);
  mleError = sqrt(pow((mlex - tx[node]),2)+pow((mley - ty[node]),2));

  //proximity
  prox = proy = 0.0;
  numPro = 0;
  for(int i=0;i<ANCHORS;i++){
    if(rssRankSequence[i] == 0){
      prox += ax[i];
      proy += ay[i];
      numPro++;
    }
  }
  prox = prox/(numPro*1.0);
  proy = proy/(numPro*1.0);
  proError = sqrt(pow((prox - tx[node]),2)+pow((proy - ty[node]),2));

  // Centroid 
  cenx = ceny = 0.0;
  numCen = 0;
  for(int i=0;i<ANCHORS;i++){
    if((rssRankSequence[i] == 0) || (rssRankSequence[i] == 1) || (rssRankSequence[i] == 2)){
      cenx += ax[i];
      ceny += ay[i];
      numCen++;
    }
  }
  cenx = cenx/(numCen*1.0);
  ceny = ceny/(numCen*1.0);
  cenError = sqrt(pow((cenx - tx[node]),2)+pow((ceny - ty[node]),2));
  
  printf("%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\t%4.2f\n",
	 tx[node],ty[node],T,kx,ky,maxKendall,kError,mlex,mley,mleError,prox,proy,proError,cenx,ceny,cenError);
  
  // House keeping
  delete EL;
  delete FL;
  delete VL;
  for(int i=0;i<4;i++){
    delete be[i];
    delete bv[i];
  }
  delete f;
  for(int i=0;i<numLines;i++)
    delete line[i];
  numRankSequences = 0;
  numBitSequences = 0;
}

