/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   "Copyright (c) 2004, 2005 The University of Southern California"
%   All rights reserved.
%
%   Permission to use, copy, modify, and distribute this script and its
%   documentation for any purpose, without fee, and without written agreement is
%   hereby granted, provided that the above copyright notice, the following
%   two paragraphs and the names in the credits appear in all copies of this software.
%
%   NO REPRESENTATIONS ARE MADE ABOUT THE SUITABILITY OF THE SCRIPT FOR ANY
%   PURPOSE. IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY. NO
%   LIABILITY IS ASSUMED BY THE DEVELOPERS.
%
%
%   Authors:  	     Kiran Yedavalli
%   Director:        Prof. Bhaskar Krishnamachari
%   Autonomous Networks Research Group, University of Southern California
%   http://ceng.usc.edu/~anrg
%
%   Contact: kyedavalli@gmail.com
%   Current Version: 1.0, 2005/11/01
%   Date last modified: 2005/11/01
%
%   Anything following a // is treated as a comment.
%
%
%   Description:
%	This program evaluates the location of unknown nodes using
%  	the Sequence-Based Localization technique. It also impliments
%  	Proximity, Maximum Likelihood Estimation (Least Squares Estimation),
%  	and Centroid techniques for comparison.
%
%	This technique is described in detail in:
%  	Kiran Yedavalli, Bhaskar Krishnamachari, "Sequence-Based Localization in Wireless Sensor Networks",
%  	To Appear in IEEE Transaction on Mobile Computing, 2007.
%
%  	We used the g++ compiler to compile the program.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <string.h>

#define MAXNUM 20000
#define epsilon (0.0001)
#define ANCHORS 10
#define RANGE 12
#define ETA (2.875)
#define TXP 5
#define PL 55
#define RES 10
#define LINES ((int)(ANCHORS*(ANCHORS-1)*0.5))

int printIt;
int bitSequence[MAXNUM][ANCHORS][ANCHORS];
int numBitSequences = 0;

int rankSequence[MAXNUM][ANCHORS];
int numRankSequences = 0;
float ax[ANCHORS],ay[ANCHORS];

struct Centroid{
  float cx,cy;
};

Centroid centroidList[10000];

class Vertex;
class Face;
class HalfEdge;
class FaceList;
class EdgeList;
class VertexList;
class Line;
float getFaceArea(Face *f);
Centroid getFaceCentroid(Face *f);

// ***** CLASSES ***** //

// Vertex Class
class Vertex{
public:
  float x,y;
  HalfEdge *incEdge;
  Vertex(float a, float b){x=a;y=b;};
};


// Face Class
class Face{
public:
  float area;
  Centroid cent;
  HalfEdge *incEdge;
};

// Half Edge Class
class HalfEdge{
public:
  Vertex *orig,*dest;
  HalfEdge(Vertex* o, Vertex* d){orig=o;dest=d;};
  Face *incFace;
  HalfEdge *twin,*next,*prev;
};

// Line class
class Line{
 public:
  float lx,ly,rx,ry;
  int ne,nv;
  Vertex *lv[MAXNUM];
  HalfEdge *le[MAXNUM];
  Line(){ne = 0;nv = 0;};
};

// List to store the face pointers
class FaceList{
 public:
  int numFaces;
  Face *fp[MAXNUM];
  int bitMap[MAXNUM];
  Centroid centroid[MAXNUM];
  float area[MAXNUM];

  FaceList();
  void addFace(Face*);
  void deleteFace(Face*);
  void printFaceList();
  float getAverageAreaOfFaces();
  float getAreaStdOfFaces();
  int getNumberOfFaces();
  float getMinFaceArea();
  float getMaxFaceArea();
  void getBitSequences();
  void getRankSequences();
};

FaceList::FaceList(){
  numFaces = 0;
  for(int i=0;i<MAXNUM;i++)
    bitMap[i] = 0;
}

void FaceList::addFace(Face *f){
  fp[numFaces] = f;
  bitMap[numFaces] = 1;
  if(printIt == 1)
    printf("FaceList: Face Pointer Added At %d.\n",numFaces);
  numFaces++;
}

void FaceList::deleteFace(Face *f){
  int test = 0;
  for(int i=0;i<MAXNUM;i++){
    if((!test) && (fp[i] == f)){
      bitMap[i] = 0;
      test = 1;
      if(printIt == 1)
	printf("FaceList: Face Pointer Found At %d - Deleting\n",i);
    }
  }
  if(test == 0){
    if(printIt == 1)
      printf("FaceList ERROR: No Such Face Exists For Deletion\n");
  }
}

void FaceList::printFaceList(){
  int count = 0;
  float avgArea = 0.0;
  float stdArea = 0.0;
  float totalArea = 0.0;
  printf("\n*** FACELIST ***\n");
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      count++;
      area[i] = getFaceArea(fp[i]);
      avgArea += area[i];
      totalArea += area[i];
      centroid[i] = getFaceCentroid(fp[i]);
      printf("%d\t%f\t(%f, %f)\n",i,area[i],centroid[i].cx,centroid[i].cy);
    }
  }
  avgArea = avgArea/(count*1.0);
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      stdArea += pow((area[i]-avgArea),2);
    }
  }
  stdArea = sqrt(stdArea/(count*1.0));
  printf("NUM-FACES: %d\tTOTAL-AREA: %f\tAVERAGE-AREA: %f\tAREA-STD: %f\n",count,totalArea,avgArea,stdArea);
}

float FaceList::getAverageAreaOfFaces(){
  float avgArea = 0.0;
  int count = 0;
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      count++;
      avgArea += getFaceArea(fp[i]);
    }
  }
  avgArea = avgArea/(count*1.0);

  return (avgArea);
}

float FaceList::getAreaStdOfFaces(){
  int count = 0;
  float avgArea = 0.0;
  float stdArea = 0.0;
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      count++;
      area[i] = getFaceArea(fp[i]);
      avgArea += area[i];
    }
  }
  avgArea = avgArea/(count*1.0);
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      stdArea += pow((area[i]-avgArea),2);
    }
  }
  stdArea = sqrt(stdArea/(count*1.0));
  return (stdArea);
}

int FaceList::getNumberOfFaces(){
  int count = 0;
  for(int i=0;i<MAXNUM;i++)
    if(bitMap[i] == 1)
      count++;
  return (count);
}

float FaceList::getMaxFaceArea(){
  float maxArea = 0.0;
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      area[i] = getFaceArea(fp[i]);
      if(area[i] > maxArea)
	maxArea = area[i];
    }
  }
  return (maxArea);
}

float FaceList::getMinFaceArea(){
  float minArea = 1000000.0;
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      area[i] = getFaceArea(fp[i]);
      if(area[i] < minArea)
	minArea = area[i];
    }
  }
  return (minArea);
}

void FaceList::getBitSequences(){
  float d[ANCHORS];
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){

      centroidList[numBitSequences] = getFaceCentroid(fp[i]);

      // Get the distances to all anchors from this centroid point
      for(int p=0;p<ANCHORS;p++)
        d[p] = sqrt(pow((centroidList[numBitSequences].cx - ax[p]),2)+pow((centroidList[numBitSequences].cy - ay[p]),2));

      // Evaluate the sequence based on these distances
      for(int j=0;j<ANCHORS;j++){
        for(int k=j+1;k<ANCHORS;k++){

          if(d[j] < d[k])
            bitSequence[numBitSequences][j][k] = 1;
          else if(d[j] > d[k])
            bitSequence[numBitSequences][j][k] = 0;
          else
            bitSequence[numBitSequences][j][k] = -1;
        }
      }

      numBitSequences++;
    }
  }
  if(printIt == 1){
    printf("\n*** FACELIST: BIT SEQUENCE LIST ***\n");
    for(int i=0;i<numBitSequences;i++){
      printf("%d\t",i);

      for(int j=0;j<ANCHORS;j++){
        for(int k=j+1;k<ANCHORS;k++){
          printf("%d",bitSequence[i][j][k]);
        }
      }

      printf("\t[%f, %f]\n",centroidList[i].cx,centroidList[i].cy);
    }
  }
}

void FaceList::getRankSequences(){
  float d[ANCHORS],minD;
  int count,minIndex;
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){

      centroidList[numRankSequences] = getFaceCentroid(fp[i]);

      // Get the distances to all anchors from this centroid point
      for(int p=0;p<ANCHORS;p++)
        d[p] = sqrt(pow((centroidList[numRankSequences].cx - ax[p]),2)
		    +pow((centroidList[numRankSequences].cy - ay[p]),2));

      // rank the anchors based on these distances
      count = 0;
      while(count < ANCHORS){
	minD = 1000000.0;
	for(int p=0;p<ANCHORS;p++){
	  if(d[p] < minD){
	    minD = d[p];
	    minIndex = p;
	  }
	}
	d[minIndex] = 1000000.0;
	rankSequence[numRankSequences][minIndex] = count;
	count++;
      }

      numRankSequences++;
    }
  }
  if(printIt == 1){
    printf("\n*** FACELIST: RANK SEQUENCE LIST ***\n");
    for(int i=0;i<numRankSequences;i++){
      printf("%d\t",i);

      for(int j=0;j<ANCHORS;j++){
	printf("%d",rankSequence[i][j]);
      }

      printf("\t[%f, %f]\n",centroidList[i].cx,centroidList[i].cy);
    }
  }
}

//-------------------------------------------------------------------

// List to store Edge Pointers
class EdgeList{
 public:
  int numEdges;
  HalfEdge *ep[MAXNUM];
  int bitMap[MAXNUM];
  EdgeList();
  void addEdge(HalfEdge*);
  void deleteEdge(HalfEdge*);
  HalfEdge* getEdgeWithOrigDest(Vertex*,Vertex*);
  int isThereEdgeWithOrigDest(Vertex*,Vertex*);
  HalfEdge* getEdgeWithPoint(float,float);
  HalfEdge* getBoundaryEdgeWithOrig(float,float);
  void printEdgeList();
  void printActualEdgeList();
  int getNumberOfEdges();
  float getAverageLengthOfEdges();
  void getBitSequences();
  void getRankSequences();
};

EdgeList::EdgeList(){
  numEdges = 0;
  for(int i=0;i<MAXNUM;i++)
    bitMap[i] = 0;
}

void EdgeList::addEdge(HalfEdge *e){
  ep[numEdges] = e;
  bitMap[numEdges] = 1;
  if(printIt == 1){
    printf("EdgeList: Edge([%f, %f]->[%f, %f]) Added At %d.\n",
	   e->orig->x,e->orig->y,e->dest->x,e->dest->y,numEdges);
  }
  numEdges++;
}

void EdgeList::deleteEdge(HalfEdge *e){
  int test = 0;
  for(int i=0;i<MAXNUM;i++){
    if((!test) && (fabs(ep[i]->orig->x - e->orig->x) < epsilon) && (fabs(ep[i]->orig->y - e->orig->y) < epsilon)
       && (fabs(ep[i]->dest->x - e->dest->x) < epsilon) && (fabs(ep[i]->dest->y - e->dest->y) < epsilon)){
      bitMap[i] = 0;
      test++;
      if(printIt == 1){
	printf("EdgeList: Edge([%f, %f]->[%f, %f]) Found At %d - Deleting\n",
	       e->orig->x,e->orig->y,e->dest->x,e->dest->y,i);
      }
    }
  }
  if(test != 1){
    if(printIt == 1)
      printf("EdgeList ERROR: No Such Edge Exists For Deletion\n");
  }
}

HalfEdge* EdgeList::getEdgeWithPoint(float a, float b){
  int test = 0;
  float d,d1,d2;
  for(int i=0;i<MAXNUM;i++){
    if((!test) && (bitMap[i] == 1)){
      d = sqrt(pow((ep[i]->orig->x - ep[i]->dest->x),2) + pow((ep[i]->orig->y - ep[i]->dest->y),2));
      d1 = sqrt(pow((ep[i]->orig->x - a),2)+pow((ep[i]->orig->y - b),2));
      d2 = sqrt(pow((ep[i]->dest->x - a),2)+pow((ep[i]->dest->y - b),2));
      if(fabs(d-(d1+d2)) < epsilon){
	test = 1;
	if(printIt == 1){
	  printf("EdgeList: Found Edge([%f, %f]->[%f, %f]) With Point[%f,%f] At %d\n",
		 ep[i]->orig->x,ep[i]->orig->y,ep[i]->dest->x,ep[i]->dest->y,a,b,i);
	}
	return ep[i];
      }
    }
  }
  if(test == 0){
    if(printIt == 1)
      printf("EdgeList: No Edge Exists With Point[%f, %f]\n",a,b);
    return NULL;
  }
}

HalfEdge* EdgeList::getBoundaryEdgeWithOrig(float a, float b){
  int test = 0;
  for(int i=0;i<MAXNUM;i++){
    if((!test) && (bitMap[i] == 1) && (fabs(ep[i]->orig->x - a) < epsilon) &&
       (fabs(ep[i]->orig->y - b) < epsilon) && (ep[i]->twin == NULL)){
      test = 1;
      if(printIt == 1){
	printf("EdgeList: Found Boundary Edge([%f, %f]->[%f, %f]) With Point[%f, %f] As The Origin",
	       ep[i]->orig->x,ep[i]->orig->y,ep[i]->dest->x,ep[i]->dest->y,a,b);
      }
      return ep[i];
    }
  }

  if(test == 0){
    if(printIt == 1)
      printf("EdgeList: No Boundary Edge Exists With The Point As The Origin\n");
    return NULL;
  }
}

HalfEdge* EdgeList::getEdgeWithOrigDest(Vertex *o, Vertex *d){
  int test = 0;
  for(int i=0;i<MAXNUM;i++){
    if((!test) && (bitMap[i] == 1) && (fabs(ep[i]->orig->x - o->x) < epsilon)
       && (fabs(ep[i]->orig->y - o->y)< epsilon) && (fabs(ep[i]->dest->x - d->x)< epsilon)
       && (fabs(ep[i]->dest->y - d->y)< epsilon)){
      test++;
      return ep[i];
    }
  }
  if(test == 0){
    if(printIt == 1){
      printf("EdgeList ERROR: No Edge Exists With ORIG(%f, %f) And DEST(%f, %f)\n",
	     o->x,o->y,d->x,d->y);
    }
    return NULL;
  }
}

int EdgeList::isThereEdgeWithOrigDest(Vertex *o, Vertex *d){
  int test = 0;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (fabs(ep[i]->orig->x - o->x)< epsilon) && (fabs(ep[i]->orig->y - o->y)< epsilon) &&
       (fabs(ep[i]->dest->x - d->x) < epsilon) && (fabs(ep[i]->dest->y - d->y)< epsilon)){
      test++;
      if(printIt == 1){
	printf("The Edge([%f, %f]->[%f, %f]) Exists In the EdgeList At %d\n",o->x,o->y,d->x,d->y,i);
      }
    }
  }
  if(test == 0){
    if(printIt == 1)
      printf("There Is No Such Edge In The EdgeList\n");
  }
  return (test);
}

int EdgeList::getNumberOfEdges(){
  int count = 0;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (ep[i]->twin != NULL)){
      //printf("EDGE %d: \t[%f, %f]->[%f, %f]\n",i,ep[i]->orig->x,ep[i]->orig->y,ep[i]->dest->x,ep[i]->dest->y);
      count++;
    }
  }
  count = (int)(count*0.5);
  return (count);
}

float EdgeList::getAverageLengthOfEdges(){
  float avgLength = 0.0;
  int count = 0;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (ep[i]->twin != NULL)){
      avgLength += sqrt(pow((ep[i]->orig->x - ep[i]->dest->x),2)+pow((ep[i]->orig->y - ep[i]->dest->y),2));
      count++;
    }
  }
  avgLength = (avgLength)/(count*1.0);
  return (avgLength);
}

void EdgeList::printEdgeList(){
  int count = 0;
  printf("\n*** EDGELIST ***\n");
  for(int i=0;i<MAXNUM;i++)
    if(bitMap[i] == 1)
      count++;
  printf("NUM-EDGES: %d\n",count);

  count = 0;
  for(int i=0;i<MAXNUM;i++){
    if(bitMap[i] == 1){
      count++;
      printf("%d\t([%3.1f,%3.1f]-[%3.1f,%3.1f])\t([%3.1f,%3.1f]-[%3.1f,%3.1f])\t([%3.1f,%3.1f]-[%3.1f,%3.1f])\t",i,
	     ep[i]->orig->x,ep[i]->orig->y,ep[i]->dest->x,ep[i]->dest->y,
	     ep[i]->next->orig->x,ep[i]->next->orig->y,ep[i]->next->dest->x,ep[i]->next->dest->y,
	     ep[i]->prev->orig->x,ep[i]->prev->orig->y,ep[i]->prev->dest->x,ep[i]->prev->dest->y);
      if(ep[i]->twin != NULL){
	printf("([%3.1f,%3.1f]-[%3.1f,%3.1f])\n",ep[i]->twin->orig->x,ep[i]->twin->orig->y,
	       ep[i]->twin->dest->x,ep[i]->twin->dest->y);
      }
      else{
	printf("\n");
      }
    }
  }
}

void EdgeList::printActualEdgeList(){
  int count = 0;
  printf("\n*** EDGELIST EXCLUDING THE BOUNDARY EDGES ***\n");
  for(int i=0;i<MAXNUM;i++)
    if((bitMap[i] == 1) && (ep[i]->twin != NULL))
      count++;
  printf("NUM-EDGES: %d\n",count);

  count = 0;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (ep[i]->twin != NULL)){
      count++;
      printf("%d\t([%3.1f,%3.1f]-[%3.1f,%3.1f])\t([%3.1f,%3.1f]-[%3.1f,%3.1f])\t([%3.1f,%3.1f]-[%3.1f,%3.1f])\t([%3.1f,%3.1f]-[%3.1f,%3.1f])\n",i,
             ep[i]->orig->x,ep[i]->orig->y,ep[i]->dest->x,ep[i]->dest->y,
             ep[i]->next->orig->x,ep[i]->next->orig->y,ep[i]->next->dest->x,ep[i]->next->dest->y,
             ep[i]->prev->orig->x,ep[i]->prev->orig->y,ep[i]->prev->dest->x,ep[i]->prev->dest->y,
             ep[i]->twin->orig->x,ep[i]->twin->orig->y,ep[i]->twin->dest->x,ep[i]->twin->dest->y);
    }
  }
}

void EdgeList::getBitSequences(){
  float d[ANCHORS];
  int check,repeat,count;
  count = numBitSequences;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (ep[i]->twin != NULL)){

      centroidList[numBitSequences].cx = (ep[i]->orig->x + ep[i]->dest->x)*0.5;
      centroidList[numBitSequences].cy = (ep[i]->orig->y + ep[i]->dest->y)*0.5;

      // Get the distances to all anchors from this centroid point
      for(int p=0;p<ANCHORS;p++)
        d[p] = sqrt(pow((centroidList[numBitSequences].cx - ax[p]),2)+pow((centroidList[numBitSequences].cy - ay[p]),2));

      // Evaluate the sequence based on these distances
      for(int j=0;j<ANCHORS;j++){
        for(int k=j+1;k<ANCHORS;k++){

          if(d[j] < d[k])
            bitSequence[numBitSequences][j][k] = 1;
          else if(d[j] > d[k])
            bitSequence[numBitSequences][j][k] = 0;
          else
            bitSequence[numBitSequences][j][k] = -1;
        }
      }

      repeat = 0;
      // Check if this is a repeat sequence
      for(int j=count;j<numBitSequences;j++){
        if(repeat == 0){
          check = 0;
          for(int p=0;p<ANCHORS;p++){
            for(int q=p+1;q<ANCHORS;q++){
              if(bitSequence[numBitSequences][p][q] == bitSequence[j][p][q])
                check++;
            }
          }
          if(check == LINES)
            repeat = 1;
        }
      }

      // If its not a repeat update the number of sequences
      if(repeat == 0)
        numBitSequences++;
    }
  }
  if(printIt == 1){
    printf("\n*** EDGELIST: BIT SEQUENCE LIST ***\n");
    for(int i=0;i<numBitSequences;i++){
      printf("%d\t",i);

      for(int j=0;j<ANCHORS;j++){
        for(int k=j+1;k<ANCHORS;k++){
          printf("%d",bitSequence[i][j][k]);
        }
      }
      printf("\t[%f, %f]\n",centroidList[i].cx,centroidList[i].cy);
    }
  }
}

void EdgeList::getRankSequences(){
  float d[ANCHORS],minD;
  int check,repeat,count;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (ep[i]->twin != NULL)){

      centroidList[numRankSequences].cx = (ep[i]->orig->x + ep[i]->dest->x)*0.5;
      centroidList[numRankSequences].cy = (ep[i]->orig->y + ep[i]->dest->y)*0.5;

      // Get the distances to all anchors from this centroid point
      for(int p=0;p<ANCHORS;p++)
        d[p] = sqrt(pow((centroidList[numRankSequences].cx - ax[p]),2)+
		    pow((centroidList[numRankSequences].cy - ay[p]),2));

      // rank the anchors based on these distances
      count = 0;
      while(count < ANCHORS){
	minD = 1000000.0;
	for(int p=0;p<ANCHORS;p++){
	  if(d[p] < minD)
	    minD = d[p];
	}
	check = 0;
	for(int p=0;p<ANCHORS;p++){
	  if(d[p] == minD){
	    d[p] = 1000000.0;
	    rankSequence[numRankSequences][p] = count;
	    check++;
	  }
	}
	count = count + check;
      }

      repeat = 0;
      // Check if this is a repeat sequence
      for(int j=0;j<numRankSequences;j++){
        if(repeat == 0){
          check = 0;
          for(int p=0;p<ANCHORS;p++){
	    if(rankSequence[numRankSequences][p] == rankSequence[j][p])
	      check++;
	  }
          if(check == ANCHORS)
            repeat = 1;
        }
      }

      // If its not a repeat update the number of sequences
      if(repeat == 0)
        numRankSequences++;
    }
  }
  if(printIt == 1){
    printf("\n*** EDGELIST: RANK SEQUENCE LIST ***\n");
    for(int i=0;i<numRankSequences;i++){
      printf("%d\t",i);

      for(int j=0;j<ANCHORS;j++){
	printf("%d",rankSequence[i][j]);
      }
      printf("\t[%f, %f]\n",centroidList[i].cx,centroidList[i].cy);
    }
  }
}

//--------------------------------------------------------

// List to store Vertices
class VertexList{
 public:
  Vertex *vp[MAXNUM];
  int bitMap[MAXNUM];
  VertexList();
  void addVertex(Vertex*);
  int isVertexNew(Vertex*);
  void printVertexList();
  void printActualVertexList();
  int isPointNew(float,float);
  int getNumberOfVerts();
  void getBitSequences();
  void getRankSequences();
};

VertexList::VertexList(){
  for(int i=0;i<MAXNUM;i++)
    bitMap[i] = 0;
}

void VertexList::addVertex(Vertex *v){
  int notDone = 1;
  for(int i=0;i<MAXNUM;i++){
    if((notDone) && (bitMap[i] == 0)){
      vp[i] = v;
      bitMap[i] = 1;
      if(printIt == 1)
	printf("VertexList: Vertex(%f, %f) Added At %d.\n",v->x,v->y,i);
      notDone = 0;
    }
  }
}

int VertexList::isVertexNew(Vertex *v){
  int test = 0;
  for(int i=0;i<MAXNUM;i++){
    if((!test) && (bitMap[i] == 1) && (fabs(vp[i]->x - v->x) < epsilon) && (fabs(vp[i]->y - v->y) < epsilon)){
      test++;
    }
  }

  if(test == 0){
    if(printIt == 1)
      printf("Vertex(%f, %f) Is New To The Vertex List\n",v->x,v->y);
    return (1);
  }
  else{
    if(printIt == 1)
      printf("Vertex(%f, %f) Is NOT New To The Vertex List\n",v->x,v->y);
    return (0);
  }
}

int VertexList::isPointNew(float a, float b){
  int test = 0;
  for(int i=0;i<MAXNUM;i++){
    if((!test) && (bitMap[i] == 1) && (fabs(vp[i]->x - a) < epsilon) && (fabs(vp[i]->y - b) < epsilon)){
      test++;
    }
  }
  if(test == 0){
    if(printIt == 1)
      printf("Point(%f, %f) Is New To The Vertex List\n",a,b);
    return (1);
  }
  else{
    if(printIt == 1)
      printf("Point(%f, %f) Is NOT New To The Vertex List\n",a,b);
    return (0);
  }
}

int VertexList::getNumberOfVerts(){
  int count = 0;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (vp[i]->x != 0.0) && (vp[i]->x != RANGE*1.0) &&
       (vp[i]->y != 0.0) && (vp[i]->y != RANGE*1.0)){
      count++;
    }
  }
  return (count);
}

void VertexList::printVertexList(){
  int count = 0;
  printf("\n*** VERTEXLIST ***\n");
   for(int i=0;i<MAXNUM;i++){
     if(bitMap[i] == 1){
       printf("%d\t[%4.3f, %4.3f]\n",i,vp[i]->x,vp[i]->y);
       count++;
     }
   }
   printf("NUM-VERTICES: %d\n",count);
}

void VertexList::printActualVertexList(){
  int count = 0;
  printf("\n*** VERTEXLIST EXCLUDING VERTICES ON THE BOUNDING BOX ***\n");
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (vp[i]->x != 0.0) && (vp[i]->x != RANGE*1.0) &&
       (vp[i]->y != 0.0) && (vp[i]->y != RANGE*1.0)){
      printf("%d\t[%4.3f, %4.3f]\n",i,vp[i]->x,vp[i]->y);
      count++;
    }
  }
  printf("NUM-VERTICES: %d\n",count);
}

void VertexList::getBitSequences(){
  int check,repeat;
  float d[ANCHORS];
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (vp[i]->x != 0.0) && (vp[i]->x != RANGE*1.0) &&
       (vp[i]->y != 0.0) && (vp[i]->y != RANGE*1.0)){

      centroidList[numBitSequences].cx = vp[i]->x;
      centroidList[numBitSequences].cy = vp[i]->y;

      // Get the distances to all anchors from this centroid point
      for(int p=0;p<ANCHORS;p++)
        d[p] = sqrt(pow((centroidList[numBitSequences].cx - ax[p]),2)+pow((centroidList[numBitSequences].cy - ay[p]),2));

      // Evaluate the sequence based on these distances
      for(int j=0;j<ANCHORS;j++){
        for(int k=j+1;k<ANCHORS;k++){

          if(d[j] < d[k])
            bitSequence[numBitSequences][j][k] = 1;
          else if(d[j] > d[k])
            bitSequence[numBitSequences][j][k] = 0;
          else
            bitSequence[numBitSequences][j][k] = -1;
        }
      }
      numBitSequences++;
    }
  }
  if(printIt == 1){
    printf("\n*** VERTEXLIST: BIT SEQUENCE LIST ***\n");
    for(int i=0;i<numBitSequences;i++){
      printf("%d\t",i);

      for(int j=0;j<ANCHORS;j++){
        for(int k=j+1;k<ANCHORS;k++){
          printf("%d",bitSequence[i][j][k]);
        }
      }
      printf("\t[%f, %f]\n",centroidList[i].cx,centroidList[i].cy);
    }
  }
}

void VertexList::getRankSequences(){
  float d[ANCHORS],minD;
  int check,repeat,count;
  for(int i=0;i<MAXNUM;i++){
    if((bitMap[i] == 1) && (vp[i]->x != 0.0) && (vp[i]->x != RANGE*1.0) &&
       (vp[i]->y != 0.0) && (vp[i]->y != RANGE*1.0)){

      centroidList[numRankSequences].cx = vp[i]->x;
      centroidList[numRankSequences].cy = vp[i]->y;

      // Get the distances to all anchors from this centroid point
      for(int p=0;p<ANCHORS;p++){
        d[p] = sqrt(pow((centroidList[numRankSequences].cx - ax[p]),2)+
		    pow((centroidList[numRankSequences].cy - ay[p]),2));
      }

      // rank the anchors based on these distances
      count = 0;
      while(count < ANCHORS){
	minD = 1000000.0;
	for(int p=0;p<ANCHORS;p++){
	  if(d[p] < minD)
	    minD = d[p];
	}
	check = 0;
	for(int p=0;p<ANCHORS;p++){
	  if(d[p] == minD){
	    d[p] = 1000000.0;
	    rankSequence[numRankSequences][p] = count;
	    check++;
	  }
	}
	count = count + check;
      }
      /*
      // Check if this is a repeat sequence
      repeat = 0;
      for(int j=0;j<numRankSequences;j++){
        if(repeat == 0){
          check = 0;
          for(int p=0;p<ANCHORS;p++){
	    if(rankSequence[numRankSequences][p] == rankSequence[j][p])
	      check++;
	  }
          if(check == ANCHORS)
            repeat = 1;
        }
      }
      */
      numRankSequences++;
    }
  }
  if(printIt == 1){
    printf("\n*** VERTEXLIST: RANK SEQUENCE LIST ***\n");
    for(int i=0;i<numRankSequences;i++){
      printf("%d\t",i);
      for(int j=0;j<ANCHORS;j++){
	printf("%d",rankSequence[i][j]);
      }
      printf("\t[%f, %f]\n",centroidList[i].cx,centroidList[i].cy);
    }
  }
}

//*************** FUNCTIONS ********************//

float getFaceArea(Face *f){
  HalfEdge *temp;
  float a[MAXNUM],b[MAXNUM];
  int count;
  float area = 0.0;
  // Getting the vertices
  temp = f->incEdge->next;
  count = 0;
  while(temp != f->incEdge){
    a[count] = temp->orig->x;
    b[count] = temp->orig->y;
    count++;
    temp = temp->next;
  }
  a[count] = temp->orig->x;
  b[count] = temp->orig->y;
  count++;
  // Calculating the area
  for(int i=0;i<count;i++)
    area += a[(i%count)]*b[((i+1)%count)] - a[((i+1)%count)]*b[(i%count)];
  area = area*0.5;
  return (area);
};

//-------------------------------

Centroid getFaceCentroid(Face *f){
  Centroid cent;
  HalfEdge *temp;
  int count;
  float a[MAXNUM],b[MAXNUM],area;
  float x=0,y=0;
  // Getting the vertices
  temp = f->incEdge->next;
  count = 0;
  while(temp != f->incEdge){
    a[count] = temp->orig->x;
    b[count] = temp->orig->y;
    count++;
    temp = temp->next;
  }
  a[count] = temp->orig->x;
  b[count] = temp->orig->y;
  count++;
  // Getting the area
  area = getFaceArea(f);
  // Calculating the Centroid
  for(int i=0;i<count;i++){
    x += (a[(i%count)] + a[((i+1)%count)])*(a[(i%count)]*b[((i+1)%count)] - a[((i+1)%count)]*b[(i%count)]);
    y += (b[(i%count)] + b[((i+1)%count)])*(a[(i%count)]*b[((i+1)%count)] - a[((i+1)%count)]*b[(i%count)]);
  }
  cent.cx = x/(6*area);
  cent.cy = y/(6*area);

  return (cent);
};

Vertex* getVertex(Line *line, HalfEdge *edge){
  Vertex *vert;
  float m1,m2,c1,c2;
  float vx,vy;

  if((fabs(edge->orig->x - edge->dest->x)< epsilon) && (fabs(line->lx - line->rx) > epsilon)){
    if(printIt == 1)
      printf("Only The Edge Is Verticle\n");
    vx = edge->orig->x;
    m1 = (line->ry - line->ly)/(line->rx - line->lx);
    c1 = (((line->ly)*(line->rx)) - ((line->ry)*(line->lx)))/(line->rx - line->lx);
    vy =  m1*(vx) + c1;
  }
  else if((fabs(edge->orig->x - edge->dest->x) > epsilon) && (fabs(line->lx - line->rx) < epsilon)){
    if(printIt == 1)
      printf("Only The Line Is Verticle\n");
    vx = line->lx;
    m2 = (edge->dest->y - edge->orig->y)/(edge->dest->x - edge->orig->x);
    c2 = (((edge->orig->y)*(edge->dest->x)) - ((edge->dest->y)*(edge->orig->x)))/(edge->dest->x - edge->orig->x);
    vy = m2*(vx)+c2;
  }
  else if((fabs(edge->orig->x - edge->dest->x) < epsilon) && (fabs(line->lx - line->rx) < epsilon)){
    if(printIt == 1)
      printf("Both The Line And The Edge are Verticle\n");
    vx = -1000.0;
    vy = -1000.0;
  }
  else{
    m1 = (line->ry - line->ly)/(line->rx - line->lx);
    c1 = (((line->ly)*(line->rx)) - ((line->ry)*(line->lx)))/(line->rx - line->lx);
    m2 = (edge->dest->y - edge->orig->y)/(edge->dest->x - edge->orig->x);
    c2 = (((edge->orig->y)*(edge->dest->x)) - ((edge->dest->y)*(edge->orig->x)))/(edge->dest->x - edge->orig->x);
    vx = (c2-c1)/(m1-m2);
    vy = (m1*c2-m2*c1)/(m1-m2);
  }

  vert = new Vertex(vx,vy);
  return (vert);
};

int isThereIntersection(Line *line,HalfEdge *e){
  float d1,d2,d;
  Vertex *v;
  v = getVertex(line,e);
  if((v->x >= 0.0) && (v->y >= 0.0)){
    d = sqrt(pow((e->orig->x - e->dest->x),2)+pow((e->orig->y - e->dest->y),2));
    d1 = sqrt(pow((e->orig->x - v->x),2)+pow((e->orig->y - v->y),2));
    d2 = sqrt(pow((e->dest->x - v->x),2)+pow((e->dest->y - v->y),2));
    if(printIt == 1)
      printf("IS-THERE-INTERSECTION: d: %f\td1: %f\td2: %f\t(d1+d2): %f\tfabs: %f\n",d,d1,d2,(d1+d2),fabs(d-(d1+d2)));
    if(fabs(d-(d1+d2)) < epsilon){
      if(printIt == 1)
	printf("There Is Intersection(%f, %f) Between The Line And The Edge\n",v->x,v->y);
      delete v;
      return 1;
    }
    else{
      if(printIt == 1)
	printf("There Is NO Intersection Between The Line And The Edge\n");
      delete v;
      return 0;
    }
  }
  else{
    if(printIt == 1)
      printf("There Is NO Intersection Between The Line And The Edge\n");
    delete v;
    return 0;
  }
};

int isVertexNewToEdge(HalfEdge *ef,Vertex *vf){
  int test;
  float ox,oy,dx,dy,vx,vy;
  ox = ef->orig->x;
  oy = ef->orig->y;
  dx = ef->dest->x;
  dy = ef->dest->y;
  vx = vf->x;
  vy = vf->y;

  if((fabs(ox - vx) < epsilon) && (fabs(oy - vy) < epsilon)){
    test = 0;
    if(printIt == 1)
      printf("Vertex(%f, %f) Is The Origin Of The Edge([%f, %f]->[%f, %f])\n",vx,vy,ox,oy,dx,dy);
  }
  else if((fabs(dx - vx) < epsilon) && (fabs(dy - vy) < epsilon)){
    test = 1;
    if(printIt == 1)
      printf("Vertex(%f, %f) Is The Destination Of The Edge([%f, %f]->[%f, %f])\n",vx,vy,ox,oy,dx,dy);
  }
  else{
    test = 2;
    if(printIt == 1)
      printf("Vertex(%f, %f) Is New To The Edge([%f, %f]->[%f, %f])\n",vx,vy,ox,oy,dx,dy);
  }

  return (test);
};

int isVertexNewToLine(Line *line,Vertex *v){
  int test = 1;
  for(int i=0;i<line->nv;i++){
    if((fabs(line->lv[i]->x - v->x) < epsilon) && (fabs(line->lv[i]->y - v->y) < epsilon)){
      if(printIt == 1)
	printf("Vertex[%f, %f] Is NOT New To The Line\n",v->x,v->y);
      test = 0;
    }
  }
  if(test == 1){
    if(printIt == 1)
      printf("Vertex[%f, %f] Is New To The Line\n",v->x,v->y);
  }
  return (test);
};

float gaussian(float sdev){
  float x1,x2,y1,y2,w;
  int dont_calculate = 0;

  if(dont_calculate){
    y1 = y2;
    dont_calculate = 0;
  }
  else{
    w = 1.0;
    while(w >= 1.0){
      x1 = 2*drand48()-1;
      x2 = 2*drand48()-1;
      w = x1*x1+x2*x2;
    }
    w = sqrt((-2*log(w))/w);
    y1 = x1*w;
    y2 = x2*w;
    dont_calculate = 1;
  }
  return(y1*sdev);
};
